-- Cross Platform Music Compiler
-- /Mic, 2008

-- This is the compiler part of XPMCK, which apart from the compiler also contains
-- playback libraries for the supported targets.

without warning

include file.e
include get.e
include wildcard.e
include machine.e
include assoc.e
include globals.e
include list.e
include string.e
include util.e
include vgm.e



-- Output all pending notes to the current song sequence			
procedure write_note(integer forceOctChange)
	atom frames, cutoffFrames, scaling
	integer length32, length256, w1, w2
	
	for i = 1 to length(supportedChannels) do
		if length(currentNote[i]) then
			if integer(tuple[i]) then
				songLen[songNum][i] += currentNote[i][2]
				frames = (updateFreq * 60.0) / currentTempo[i]			-- frames per quarternote

				if useFractionalDelays then
					length32 = floor((frames / 8) * 256)
					scaling = 256
					--frames *= currentNote[i][2] / 8
					frames = length32 * currentNote[i][2]
					if currentCutoff[i] < 0 then
						cutoffFrames = -currentCutoff[i] * scaling
						if cutoffFrames > frames then
							cutoffFrames = frames
						end if
					else
						cutoffFrames = (frames * (8 - currentCutoff[i])) / 8
					end if
					if currentNote[i][1] != rest then
						frames -= cutoffFrames
					else
						cutoffFrames = 0
					end if
					if frames < 0 then
						ERROR("Note has negative length", lineNum)
					end if
				else
					length32 = floor(frames / 8)				-- frames per 1/32 note
					frames = length32 * currentNote[i][2]
					if currentCutoff[i] < 0 then
						cutoffFrames = -currentCutoff[i]
						if cutoffFrames > frames then
							cutoffFrames = frames
						end if
					else				
						cutoffFrames = floor((frames * (8 - currentCutoff[i])) / 8)
					end if
					frames -= cutoffFrames --floor((frames * currentCutoff[i]) / 8)
				end if

				if frames < 1.0 * scaling then
					WARNING("Note is too short, will not be heard", lineNum)
				elsif floor(frames) > #3FFF * scaling + (scaling - 1) then
					WARNING("Note is too long, cutting at 16383 frames", lineNum)
					frames = #3FFF * scaling + (scaling - 1)
				end if

				if currentNote[i][1] = rest then
					currentNote[i][1] = CMD_REST
				else
					currentNote[i][1] = remainder(currentNote[i][1], 12)
				end if

				if pendingOctChange[i] = 1 then
					currentNote[i][1] = or_bits(currentNote[i][1], CMD_OCTUP)
				elsif pendingOctChange[i] = -1 then
					currentNote[i][1] = or_bits(currentNote[i][1], CMD_OCTDN)
				end if

				if useFractionalDelays then
					w1 = floor(frames) 
					w2 = floor(cutoffFrames) 
					if floor(frames) > 127 * scaling then
						songs[songNum][i] &= {currentNote[i][1],
								      or_bits(floor(w1 / #8000), #80),
								      and_bits(floor(w1 / #100), #7F),
								      and_bits(w1, #FF)}
					else
						songs[songNum][i] &= {currentNote[i][1],
								      floor(w1 / #100),
								      and_bits(w1, #FF)}
					end if
					if floor(cutoffFrames) > 127 * scaling then
						songs[songNum][i] &= {CMD_REST,
								      or_bits(floor(w2 / #8000), #80),
								      and_bits(floor(w2 / #100), #7F),
								      and_bits(w2, #FF)}
					elsif w2 >= 1 then
						songs[songNum][i] &= {CMD_REST, 
								      floor(w2 / #100),
								      and_bits(w2, #FF)}

					end if
				else
					w1 = floor(frames)
					w2 = floor(cutoffFrames)

					if w1 > 127 then
						songs[songNum][i] &= {currentNote[i][1],
								      or_bits(floor(w1 / 128), #80),
								      and_bits(w1, #7F)}
					elsif w1 >= 1 then
						songs[songNum][i] &= {currentNote[i][1], w1}
					end if
					if w2 > 127 then
						songs[songNum][i] &= {CMD_REST,
								      or_bits(floor(w2 / 128), #80),
								      and_bits(w2, #7F)}
					elsif w2 >= 1 then
						songs[songNum][i] &= {CMD_REST, w2}
					end if
				end if
			else
				if pendingOctChange[i] = 1 then
					currentNote[i][1] = or_bits(currentNote[i][1], CMD_OCTUP)
				elsif pendingOctChange[i] = -1 then
					currentNote[i][1] = or_bits(currentNote[i][1], CMD_OCTDN)
				end if
				tuple[i] = append(tuple[i], currentNote[i])
			end if
		
			currentNote[i] = {}
			pendingOctChange[i] = 0
			hasAnyNote[i] = 1
			
		elsif forceOctChange and pendingOctChange[i] then
			songs[songNum][i] &= or_bits(CMD_OCTAVE, currentOctave[i] - minOctave[i])
			pendingOctChange[i] = 0
		end if
	end for
end procedure


procedure write_tuple(integer chn, sequence t, integer len)
	atom totalFrames, frames, cutoffFrames, scaling
	integer length32, w1, w2

	songLen[songNum][chn] += len
	totalFrames = (updateFreq * 60.0) / currentTempo[chn]
	if useFractionalDelays then
		length32 = floor((totalFrames / 8) * 256)
		scaling = 256
		totalFrames = length32 * len 	
		--totalFrames *= len / 8

		if currentCutoff[chn] < 0 then
			cutoffFrames = -currentCutoff[chn] * scaling
			if cutoffFrames > totalFrames then
				cutoffFrames = totalFrames
			end if
		else
			cutoffFrames = (totalFrames * (8 - currentCutoff[chn])) / 8
		end if
		totalFrames -= cutoffFrames
		if totalFrames < 0 then
			ERROR("Tuple has negative length", lineNum)
		end if
	else
	end if
	
	if totalFrames < 1.0 * scaling then
		WARNING("Note is too short, will not be heard", lineNum)
	elsif floor(totalFrames) > #3FFF * scaling + (scaling - 1) then
		WARNING("Note is too long, cutting at 16383 frames", lineNum)
		totalFrames = #3FFF * scaling + (scaling - 1)
	end if

	frames = totalFrames / length(t)
	w1 = floor(frames) -- * 256)
	for i = 1 to length(t) do
		t[i][2] = w1
	end for
	for i = 1 to floor(totalFrames) - (w1 * length(t)) do
		t[remainder(i, length(t)) + 1][2] += 1
	end for
	w2 = floor(cutoffFrames) -- * 256)
	
	for i = 1 to length(t) do
		if t[i][1] = rest then
			t[i][1] = CMD_REST
		else
			t[i][1] = remainder(t[i][1], 12)
		end if

		if useFractionalDelays then
			if floor(t[i][2] / 256) > 127 then
				songs[songNum][chn] &= {t[i][1],
						        or_bits(floor(t[i][2] / #8000), #80),
						        and_bits(floor(t[i][2] / #100), #7F),
						        and_bits(t[i][2], #FF)}
			else
				songs[songNum][chn] &= {t[i][1], floor(t[i][2] / #100), and_bits(t[i][2], #FF)}
			end if
		else
		end if
	end for
	

	if floor(cutoffFrames) > 127 * scaling then
		songs[songNum][chn] &= {CMD_REST,
				      or_bits(floor(w2 / #8000), #80),
				      and_bits(floor(w2 / #100), #7F),
				      and_bits(w2, #FF)}
	elsif w2 >= 1 then
		songs[songNum][chn] &= {CMD_REST, 
				      floor(w2 / #100),
				      and_bits(w2, #FF)}
	end if
end procedure



-- Extract the sample data from a WAV file and convert it to mono with a given sample rate and volume
-- UNFINISHED
function convert_wav(sequence fname, integer sampleRate, integer volume)
	integer fn
	integer wavFormat, wavChannels, wavSamplesPerSec, wavAvgBytesPerSec,
	        wavBlockAlign, wavBitsPerSample
	sequence s, wavData
	atom dw, dataStart, dataSize
	
	fn = open(fname,"rb")
	if fn < 0 then
		ERROR("Unable to open file: " & fname, lineNum)
	end if

	s = get_bytes(fn, 4)
	if not equal(s, "RIFF") then
		ERROR("No RIFF tag found in " & fname, -1)
	end if
	
	if get_dword(fn) then end if
	
	s = get_bytes(fn,4)
	if not equal(s, "WAVE") then
		ERROR("No WAVE tag found in " & fname, -1)
	end if

	dataSize = -1
	wavData = {}
	
	-- Read the chunks
	while 1 do
		-- Get the chunk ID
		s = get_bytes(fn, 4)
		if length(s) < 4 then
			exit
		end if

		-- Get the chunk size
		dw = get_dword(fn)
	
		if equal(s, "fmt ") then
			wavFormat = get_word(fn)
			wavChannels = get_word(fn)
			wavSamplesPerSec = get_dword(fn)
			wavAvgBytesPerSec = get_dword(fn)
			wavBlockAlign = get_word(fn)
			wavBitsPerSample = get_word(fn)
		
			if wavFormat!=1 or (wavBitsPerSample!=8 and wavBitsPerSample!=16) or wavChannels > 2 then
				ERROR("Unsupported wav format in " & fname, -1)
			end if
		
		elsif equal(s, "data") then
			dataStart = where(fn)
			dataSize = dw
		end if
	end while
	
	return wavData
end function


-- Output table data in WLA-DX format 
function output_wla_table(sequence name, sequence tbl, integer canLoop, integer scaling)
	integer bytesWritten
	
	bytesWritten = 0
	
	if length(tbl[1]) then
		for i = 1 to length(tbl[1]) do
			printf(outFile, name & "_%d:", tbl[1][i])
			for j = 1 to length(tbl[2][i][2]) do
				if canLoop and j = length(tbl[2][i][2]) and length(tbl[2][i][3]) = 0 then
					if j > 1 then
						puts(outFile, ", $80")
					end if
					printf(outFile, {13, 10} & name & "_%d_loop:" & {13, 10}, tbl[1][i])
					printf(outFile, ".db $%02x, $80", and_bits(tbl[2][i][2][j] * scaling, #FF))
					bytesWritten += 3
				elsif j = 1 then
					printf(outFile, {13, 10} & ".db $%02x", and_bits(tbl[2][i][2][j] * scaling, #FF))
					bytesWritten += 1
				else
					printf(outFile, ", $%02x", and_bits(tbl[2][i][2][j] * scaling, #FF))
					bytesWritten += 1
				end if
			end for
			if canLoop and length(tbl[2][i][3]) then
				if length(tbl[2][i][2]) then
					puts(outFile, ", $80")
					bytesWritten += 1
				end if
				printf(outFile, {13, 10} & name & "_%d_loop:" & {13, 10}, tbl[1][i])
				for j = 1 to length(tbl[2][i][3]) do
					if j = 1 then
						printf(outFile, ".db $%02x", and_bits(tbl[2][i][3][j] * scaling, #FF))
					else
						printf(outFile, ", $%02x", and_bits(tbl[2][i][3][j] * scaling, #FF))
					end if
					bytesWritten += 1
				end for
				puts(outFile, ", 128")
				bytesWritten += 1
			end if
			puts(outFile, {13, 10})
		end for
		puts(outFile, name & "_tbl:" & {13, 10})
		for i = 1 to length(tbl[1]) do
			printf(outFile, ".dw " & name & "_%d" & {13, 10}, tbl[1][i])
			bytesWritten += 2
		end for
		if canLoop then
			puts(outFile, name & "_loop_tbl:" & {13, 10})
			for i = 1 to length(tbl[1]) do
				printf(outFile, ".dw " & name & "_%d_loop" & {13, 10}, tbl[1][i])
				bytesWritten += 2
			end for
		end if
		puts(outFile, {13, 10})
	else
		puts(outFile, name & "_tbl:" & {13, 10})
		if canLoop then
			puts(outFile, name & "_loop_tbl:" & {13, 10})
		end if
		puts(outFile, {13, 10})
	end if
	return bytesWritten
end function


-- Output table data in m68k-as format 
function output_m68kas_table(sequence name, sequence tbl, integer canLoop, integer scaling)
	integer bytesWritten
	
	bytesWritten = 0
	
	if length(tbl[1]) then
		for i = 1 to length(tbl[1]) do
			--printf(outFile, ".globl " & name & "_%d", tbl[1][i])
			printf(outFile, name & "_%d:", tbl[1][i])
			for j = 1 to length(tbl[2][i][2]) do
				if canLoop and j = length(tbl[2][i][2]) and length(tbl[2][i][3]) = 0 then
					if j > 1 then
						puts(outFile, ", 0x80")
					end if
					--printf(outFile, CRLF & ".globl " & name & "_%d_loop" & CRLF, tbl[1][i])
					printf(outFile, CRLF & name & "_%d_loop:" & CRLF, tbl[1][i])
					printf(outFile, "dc.b 0x%02x, 0x80", and_bits(tbl[2][i][2][j] * scaling, #FF))
					bytesWritten += 3
				elsif j = 1 then
					printf(outFile, CRLF & "dc.b 0x%02x", and_bits(tbl[2][i][2][j] * scaling, #FF))
					bytesWritten += 1
				else
					printf(outFile, ", 0x%02x", and_bits(tbl[2][i][2][j] * scaling, #FF))
					bytesWritten += 1
				end if
			end for
			if canLoop and length(tbl[2][i][3]) then
				if length(tbl[2][i][2]) then
					puts(outFile, ", 0x80")
					bytesWritten += 1
				end if
				--printf(outFile, {13, 10} & ".globl " & name & "_%d_loop" & CRLF, tbl[1][i])
				printf(outFile, {13, 10} & name & "_%d_loop:" & CRLF, tbl[1][i])
				for j = 1 to length(tbl[2][i][3]) do
					if j = 1 then
						printf(outFile, "dc.b 0x%02x", and_bits(tbl[2][i][3][j] * scaling, #FF))
					else
						printf(outFile, ", 0x%02x", and_bits(tbl[2][i][3][j] * scaling, #FF))
					end if
					bytesWritten += 1
				end for
				puts(outFile, ", 128")
				bytesWritten += 1
			end if
			puts(outFile, CRLF)
		end for
		puts(outFile, ".globl " & name & "_tbl" & CRLF)
		puts(outFile, name & "_tbl:" & CRLF)
		for i = 1 to length(tbl[1]) do
			printf(outFile, "dc.w " & name & "_%d" & CRLF, tbl[1][i])
			bytesWritten += 2
		end for
		puts(outFile, "dc.w 0" & CRLF)
		bytesWritten += 2
		if canLoop then
			puts(outFile, ".globl " & name & "_loop_tbl" & CRLF)
			puts(outFile, name & "_loop_tbl:" & CRLF)
			for i = 1 to length(tbl[1]) do
				printf(outFile, "dc.w " & name & "_%d_loop" & CRLF, tbl[1][i])
				bytesWritten += 2
			end for
			puts(outFile, "dc.w 0" & CRLF)
			bytesWritten += 2
		end if
		puts(outFile, {13, 10})
	else
		puts(outFile, ".globl " & name & "_tbl" & {13, 10})
		puts(outFile, name & "_tbl:" & {13, 10})
		if canLoop then
			puts(outFile, ".globl " & name & "_loop_tbl" & CRLF)
			puts(outFile, name & "_loop_tbl:" & CRLF)
		end if
		puts(outFile, "dc.w 0" & CRLF)
		bytesWritten += 2
		puts(outFile, CRLF)
	end if
	return bytesWritten
end function


procedure output_c_table(sequence name, sequence tbl, integer canLoop)
	if length(tbl[1]) then
		for i = 1 to length(tbl[1]) do
			printf(outFile, "const unsigned char " & name & "_%d[] =" & {13, 10} & "{" & {13, 10}, tbl[1][i])
			for j = 1 to length(tbl[2][i][2]) do
				if canLoop and j = length(tbl[2][i][2]) and length(tbl[2][i][3]) = 0 then
					if j > 1 then
						puts(outFile, ", 128")
					end if
					printf(outFile, "};" & {13, 10} & "const unsigned char " & name & "_%d_loop[] =" & {13, 10} & "{" & {13, 10}, tbl[1][i])
					printf(outFile, "%d,128};", tbl[2][i][2][j])
				elsif j = 1 then
					printf(outFile, "%d", tbl[2][i][2][j])
				else
					printf(outFile, ",%d", tbl[2][i][2][j])
				end if
			end for
			if canLoop and length(tbl[2][i][3]) then
				if length(tbl[2][i][2]) then
					puts(outFile, ",128};")
				end if
				printf(outFile, {13, 10} & "const unsigned char " & name & "_%d_loop[] =" & {13, 10} & "{" & {13, 10}, tbl[1][i])
				for j = 1 to length(tbl[2][i][3]) do
					if j = 1 then
						printf(outFile, "%d", tbl[2][i][3][j])
					else
						printf(outFile, ",%d", tbl[2][i][3][j])
					end if
				end for
				puts(outFile, ",128};")
			end if
			puts(outFile, {13, 10})
		end for
		puts(outFile, "const unsigned char *" & name & "_tbl[] =" & {13, 10} & "{" & {13, 10})
		for i = 1 to length(tbl[1]) do
			printf(outFile, name & "_%d", tbl[1][i])
			if i < length(tbl[1]) then
				puts(outFile, ",")
			end if
			puts(outFile, {13, 10})
		end for
		puts(outFile, "};" & {13, 10})
		if canLoop then
			puts(outFile, "const unsigned char *" & name & "_loop_tbl[] = " & {13, 10} & "{" & {13, 10})
			for i = 1 to length(tbl[1]) do
				printf(outFile, name & "_%d_loop", tbl[1][i])
				if i < length(tbl[1]) then
					puts(outFile, ",")
				end if
				puts(outFile, {13, 10})
			end for
			puts(outFile, "};")
		end if
		puts(outFile, {13, 10})
	else
		puts(outFile, "const unsigned char **" & name & "_tbl;" & {13, 10})
		if canLoop then
			puts(outFile, "const unsigned char **" & name & "_loop_tbl;" & {13, 10})
		end if
		puts(outFile, {13, 10})
	end if
end procedure



-- Output data suitable for the SEGA Master System / Game Gear playback library (WLA-DX)
procedure output_sms(integer writeVGM)
	atom factor
	integer f, machineSpeed, tableSize, cbSize, songSize
	sequence freqTbl, oct1, fileEnding, s
	
	if writeVGM then
		fileEnding = ".vgm"
	else
		fileEnding = ".asm"
	end if

	if writeVGM then
		write_vgm(shortFilename & fileEnding, 1, 1, supportsPAL, 0)
		
	else
		outFile = open(shortFilename & fileEnding, "wb")
		if outFile = -1 then
			ERROR("Unable to open file: " & shortFilename & fileEnding, -1)
		end if

		s = date()
		printf(outFile, "; Written by XPMC at %02d:%02d:%02d on " & WEEKDAYS[s[7]] & " " & MONTHS[s[2]] & " %d, %d." & {13, 10, 13, 10},
		       s[4..6] & {s[3], s[1] + 1900})
		       
		if not supportsPAL then
			puts(outFile, ".DEFINE XPMP_GAME_GEAR" & {13, 10})
		end if

		if updateFreq = 50 then
			puts(outFile, ".DEFINE XPMP_50_HZ" & {13, 10})
			machineSpeed = 3546893
		else
			machineSpeed = 3579545
		end if

		if target = TARGET_SMS and tune then
			puts(outFile, ".DEFINE XPMP_TUNE_SMS" & CRLF)
		end if

		if sum(assoc_get_references(volumeMacros)) = 0 then --not length(volumeMacros[1]) then
			puts(outFile, ".DEFINE XPMP_VMAC_NOT_USED" & CRLF)
		end if
		if sum(assoc_get_references(pitchMacros)) = 0 then
			puts(outFile, ".DEFINE XPMP_EPMAC_NOT_USED" & CRLF)
		end if
		if sum(assoc_get_references(vibratos)) = 0 then
			puts(outFile, ".DEFINE XPMP_MPMAC_NOT_USED" & CRLF)
		end if
		if not usesEN[1] then
			puts(outFile, ".DEFINE XPMP_ENMAC_NOT_USED" & CRLF)
		end if
		if not usesEN[2] then
			puts(outFile, ".DEFINE XPMP_EN2MAC_NOT_USED" & CRLF)
		end if
		
		-- Write frequency table
		--factor = 2.0
		--for i = 2 to 7 do
		--	puts(outFile, ".dw ")
		--	for n = 1 to 12 do
		--		f = floor(machineSpeed / (OCTAVE1_ALT_SMS[n] * factor * 32))
		--		printf(outFile, "$%04x", and_bits(f, #3F0) * 16 + and_bits(f, #0F) + #80)
		--		if n < 12 then
		--			puts(outFile, ",")
		--		end if
		--	end for
		--	puts(outFile, {13,10})
		--	factor *= 2.0
		--end for
		--puts(outFile, {13, 10})

		tableSize  = output_wla_table("xpmp_dt_mac", dutyMacros,   1, 1)
		tableSize += output_wla_table("xpmp_v_mac",  volumeMacros, 1, 1)
		tableSize += output_wla_table("xpmp_VS_mac", volumeSlides, 1, 1)
		tableSize += output_wla_table("xpmp_EP_mac", pitchMacros,  1, 1)
		tableSize += output_wla_table("xpmp_EN_mac", arpeggios,    1, 1)
		tableSize += output_wla_table("xpmp_MP_mac", vibratos,     0, 1)
		tableSize += output_wla_table("xpmp_CS_mac", panMacros,    1, 1)

		if verbose then
			printf(1, "Size of effect tables: %d bytes\n", tableSize)
		end if

		cbSize = 0
		puts(outFile, "xpmp_callback_tbl:" & {13, 10})
		for i = 1 to length(callbacks) do
			puts(outFile, ".dw " & callbacks[i] & {13, 10})
			cbSize += 2
		end for
		puts(outFile, {13, 10})

		if verbose then
			printf(1, "Size of callback table: %d bytes\n", cbSize)
		end if

		songSize = 0
		for n = 1 to length(songs) do
			if sequence(songs[n]) then
				for i = 1 to length(supportedChannels) do
					printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i] & ":", n)
					for j = 1 to length(songs[n][i]) do
						if remainder(j, 16) = 1 then
							puts(outFile, CRLF & ".db ")
						end if				
						printf(outFile, "$%02x", and_bits(songs[n][i][j], #FF))
						if j < length(songs[n][i]) and remainder(j, 16) != 0 then
							puts(outFile, ",")
						end if

					end for
					puts(outFile, CRLF)
					printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), songLen[n][i], songLoopLen[n][i]})
					songSize += length(songs[n][i])
				end for
			end if
		end for

		puts(outFile, {13, 10} & "xpmp_song_tbl:" & CRLF)
		for n = 1 to length(songs) do
			if sequence(songs[n]) then
				for i = 1 to length(supportedChannels) do
					printf(outFile, ".dw xpmp_s%d_channel_" & supportedChannels[i] & {13, 10}, n)
					songSize += 2
				end for
			end if
		end for

		if verbose then
			printf(1, "Total size of song(s): %d bytes\n", songSize + tableSize + cbSize)
		end if

		close(outFile)
	end if
end procedure


procedure output_gen(integer writeVGM)
	atom factor
	integer f, machineSpeed, tableSize, cbSize, songSize
	sequence freqTbl, oct1, fileEnding, s, e
	
	if writeVGM then
		fileEnding = ".vgm"
	else
		fileEnding = ".s"
	end if

	for i = 1 to length(adsrs[2]) do
		s = {0, 0, 0, 0}
		e = adsrs[2][i][2]
		s[1] = e[1]
		s[2] = e[2]
		s[3] = e[3]
		s[4] = floor(e[5] / 2) + floor(xor_bits(e[4], 31) / 2) * #10
		adsrs[2][i][2] = s
	end for
	
	for i = 1 to length(mods[2]) do
		s = mods[2][i][2]
		s[2] = s[2] * 8 + s[3]
		mods[2][i][2] = s[1..2]
	end for
	
	if writeVGM then
		write_vgm(shortFilename & fileEnding, 1, 1, 0, 1)
		
	else
		outFile = open(shortFilename & fileEnding, "wb")
		if outFile = -1 then
			ERROR("Unable to open file: " & shortFilename & fileEnding, -1)
		end if

		s = date()
		printf(outFile, "# Written by XPMC at %02d:%02d:%02d on " & WEEKDAYS[s[7]] & " " & MONTHS[s[2]] & " %d, %d." & CRLF & CRLF,
		       s[4..6] & {s[3], s[1] + 1900})
		       
		if updateFreq = 50 then
			puts(outFile, ".equ XPMP_50_HZ, 1" & CRLF)
			machineSpeed = 3546893
		else
			machineSpeed = 3579545
		end if
		
		tableSize  = output_m68kas_table("xpmp_dt_mac", dutyMacros,   1, 1)
		tableSize += output_m68kas_table("xpmp_v_mac",  volumeMacros, 1, 1)
		tableSize += output_m68kas_table("xpmp_VS_mac", volumeSlides, 1, 1)
		tableSize += output_m68kas_table("xpmp_EP_mac", pitchMacros,  1, 1)
		tableSize += output_m68kas_table("xpmp_EN_mac", arpeggios,    1, 1)
		tableSize += output_m68kas_table("xpmp_MP_mac", vibratos,     0, 1)
		tableSize += output_m68kas_table("xpmp_CS_mac", panMacros,    1, 1)
		tableSize += output_m68kas_table("xpmp_ADSR",   adsrs,        0, 1)
		tableSize += output_m68kas_table("xpmp_MOD",    mods,         0, 1)

		cbSize = 0
		
		if verbose then
			printf(1, "Size of effect tables: %d bytes\n", tableSize)
		end if

		songSize = 0
		for n = 1 to length(songs) do
			if sequence(songs[n]) then
				for i = 1 to length(supportedChannels) do
					printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i] & ":", n)
					for j = 1 to length(songs[n][i]) do
						if remainder(j, 16) = 1 then
							puts(outFile, CRLF & "dc.b ")
						end if				
						printf(outFile, "0x%02x", and_bits(songs[n][i][j], #FF))
						if j < length(songs[n][i]) and remainder(j, 16) != 0 then
							puts(outFile, ",")
						end if

					end for
					puts(outFile, CRLF)
					printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), songLen[n][i], songLoopLen[n][i]})
					songSize += length(songs[n][i])
				end for
			end if
		end for

		puts(outFile, CRLF & ".globl xpmp_song_tbl")
		puts(outFile, CRLF & "xpmp_song_tbl:" & CRLF)
		for n = 1 to length(songs) do
			if sequence(songs[n]) then
				for i = 1 to length(supportedChannels) do
					printf(outFile, "dc.w xpmp_s%d_channel_" & supportedChannels[i] & CRLF, n)
					songSize += 2
				end for
			end if
		end for
		puts(outFile, "dc.w 0" & CRLF)
		songSize += 2
		
		if verbose then
			printf(1, "Total size of song(s): %d bytes\n", songSize + tableSize + cbSize)
		end if
		
		close(outFile)
	end if
end procedure


-- Output data suitable for the Gameboy / Gameboy Color
procedure output_gbc()
	atom factor,f2,r2,smallestDiff
	integer f, tableSize
	sequence closest

	outFile = open(shortFilename & ".asm", "wb")
	if outFile = -1 then
		ERROR("Unable to open file: " & shortFilename & ".asm", -1)
	end if

	tableSize  = output_wla_table("xpmp_dt_mac", dutyMacros,   1, 1)
	tableSize += output_wla_table("xpmp_v_mac",  volumeMacros, 1, 1)
	tableSize += output_wla_table("xpmp_VS_mac", volumeSlides, 1, 1)
	tableSize += output_wla_table("xpmp_EP_mac", pitchMacros,  1, 1)
	tableSize += output_wla_table("xpmp_EN_mac", arpeggios,    1, 1)
	tableSize += output_wla_table("xpmp_MP_mac", vibratos,     0, 1)
	
	puts(outFile, "xpmp_waveform_data:")
	for i = 1 to length(waveforms[2]) do
		for j = 1 to length(waveforms[2][i]) by 2 do
			if j = 1 then
				puts(outFile, CRLF & ".db ")
			end if				
			printf(outFile, "$%02x", waveforms[2][i][j]*#10 + waveforms[2][i][j+1])
			if j < length(waveforms[2][i]) - 1 then
				puts(outFile, ",")
			end if

		end for
	end for
	puts(outFile, {13, 10, 13, 10})
				
	puts(outFile, "xpmp_callback_tbl:" & CRLF)
	for i = 1 to length(callbacks) do
		puts(outFile, ".dw " & callbacks[i] & CRLF)
	end for
	puts(outFile, {13, 10})
	
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels) do
				printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i] & ":", n)
				for j = 1 to length(songs[n][i]) do
					if remainder(j, 16) = 1 then
						puts(outFile, CRLF & ".db ")
					end if				
					printf(outFile, "$%02x", and_bits(songs[n][i][j], #FF))
					if j < length(songs[n][i]) and remainder(j, 16) != 0 then
						puts(outFile, ",")
					end if

				end for
				puts(outFile, CRLF)
				printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), songLen[n][i], songLoopLen[n][i]})
			end for
		end if
	end for
	
	puts(outFile, {13, 10} & "xpmp_song_tbl:" & CRLF)
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels) do
				printf(outFile, ".dw xpmp_s%d_channel_" & supportedChannels[i] & {13, 10}, n)
			end for
		end if
	end for
	
	close(outFile)
end procedure


procedure output_nds()
	integer first
	
	outFile = open(shortFilename & ".h", "wb")
	if outFile = -1 then
		ERROR("Unable to open file: " & shortFilename & ".h", -1)
	end if

	output_c_table("xpmp_dt_mac", dutyMacros,   1)
	output_c_table("xpmp_v_mac",  volumeMacros, 1)
	output_c_table("xpmp_VS_mac", volumeSlides, 1)
	output_c_table("xpmp_EP_mac", pitchMacros,  1)
	output_c_table("xpmp_EN_mac", arpeggios,    1)
	output_c_table("xpmp_MP_mac", vibratos,     0)
	
	puts(outFile, "const xpmp_callback xpmp_callback_tbl[] =" & {13, 10} & "{" & {13, 10})
	for i = 1 to length(callbacks) do
		puts(outFile, callbacks[i])
		puts(outFile, ",")
		puts(outFile, {13, 10})
	end for
	puts(outFile, "NULL" & {13, 10} & "};" & {13, 10})
	
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels) do
				printf(outFile, "const unsigned char xpmp_s%d_channel_" & supportedChannels[i] & " = " & {13,10} & "{" & {13, 10}, n)
				for j = 1 to length(songs[n][i]) do
					printf(outFile, "0x%02x", songs[n][i][j])
					if j < length(songs[n][i]) then
						puts(outFile, ",")
					end if
					if remainder(j, 16) = 0 then
						puts(outFile, {13, 10})
					end if
				end for
				puts(outFile, "};" & {13, 10})
				printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), songLen[n][i], songLoopLen[n][i]})
			end for
		end if
	end for
	
	puts(outFile, {13, 10} & "unsigned char *xpmp_song_tbl[] = " & {13, 10} & "{" & {13, 10})
	first = 1
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels) do
				if first then
					first = 0
				else
					puts(outFile, "," & {13,10})
				end if
				printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i], n)
			end for
		end if
	end for
	puts(outFile, {13, 10} & "};" & {13, 10})
	
	close(outFile)
end procedure


procedure compile(sequence fileNames)
	integer idx, l, m, n, n2, numChannels,
	        lastWasChannelSelect, prevLine, bypass,
	        note, firstNote, flatSharp, len, dot, tie, slur,
	        hasTie, hasSlur, writeVGM
	sequence s, t, u, macro, dontCompile, hasElse
	object o
	
	songTitle 	= "Unknown"
	songComposer 	= "Unknown"
	songProgrammer 	= "Unknown"
	
	supportsPAL 	= 0
	useFractionalDelays = 1
	writeVGM = 0
	adsrLen = 4
	adsrMax = 15
	defines = {{}, {}}
	
	tune = 0
	
	if target = TARGET_SMS then
		define("SMS", 1)
		supportedChannels 	= {'A', 'B', 'C', 'D', 'E', 'F', 'G'}
		supportsDutyChange 	= {-1,  -1,  -1,   1,  15,  15,  15}
		supportsVolumeChange 	= { 1,   1,   1,   1,   1,   1,   1}
		supportsFM		= { 0,   0,   0,   0,   1,   1,   1}
		supportsWave 		= { 0,   0,   0,   0,   0,   0,   0}
		supportsPCM 		= { 0,   0,   0,   0,   0,   0,   0}
		activeChannels 		= { 0,   0,   0,   0,   0,   0,   0}
		supportsHwToneEnv	= { 0,   0,   0,   0,   0,   0,   0}
		supportsHwVolEnv	= { 0,   0,   0,   0,   0,   0,   0}
		supportsDetune		= { 1,   1,   1,   0,   1,   1,   1}
		maxTempo 		= 250
		minOctave 		= {2, 2, 2, 2, 1, 1, 1}
		maxOctave 		= {7, 7, 7, 7, 7, 7, 7}
		minVolume 		= 0
		maxVolume 		= {15, 15, 15, 15, 15, 15, 15}
		supportsPan 		= 0
		minNote 		= {10, 10, 10, 10, 1, 1, 1}	
		maxLoopDepth 		= 2
		supportsPAL 		= 1
	
	elsif target = TARGET_SGG then
		define("SGG", 1)
		supportedChannels 	= {'A', 'B', 'C', 'D'}
		supportsDutyChange 	= {-1,  -1,  -1,   1}
		supportsVolumeChange 	= { 1,   1,   1,   1}
		supportsFM		= { 0,   0,   0,   0}
		supportsWave 		= { 0,   0,   0,   0}
		supportsPCM 		= { 0,   0,   0,   0}
		activeChannels 		= { 0,   0,   0,   0}
		supportsHwToneEnv	= { 0,   0,   0,   0}
		supportsHwVolEnv	= { 0,   0,   0,   0}
		supportsDetune		= { 1,   1,   1,   0}
		maxTempo 		= 250
		minOctave 		= {2, 2, 2, 2}
		maxOctave 		= {7, 7, 7, 7}
		minVolume 		= 0
		maxVolume 		= {15, 15, 15, 15}
		supportsPan 		= 1
		minNote 		= {10, 10, 10, 10}	-- Lowest possible note is A in octave 2
		maxLoopDepth 		= 2

	elsif target = TARGET_SMD then
		define("GEN", 1)
		define("SMD", 1)
		supportedChannels 	= {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J'}
		supportsDutyChange 	= {-1,  -1,  -1,   1,   7,   7,   7,   7,   7,   7}
		supportsVolumeChange 	= { 1,   1,   1,   1,   1,   1,   1,   1,   1,   1}
		supportsFM		= { 0,   0,   0,   0,   1,   1,   1,   1,   1,   1}
		supportsWave 		= { 0,   0,   0,   0,   0,   0,   0,   0,   0,   0}
		supportsPCM 		= { 0,   0,   0,   0,   0,   0,   0,   0,   0,   0}
		activeChannels 		= { 0,   0,   0,   0,   0,   0,   0,   0,   0,   0}
		supportsHwToneEnv	= { 0,   0,   0,   0,   0,   0,   0,   0,   0,   0}
		supportsHwVolEnv	= { 0,   0,   0,   0,   0,   0,   0,   0,   0,   0}
		supportsDetune		= { 1,   1,   1,   0,   1,   1,   1,   1,   1,   1}
		maxTempo 		= 250
		minOctave 		= {2, 2, 2, 2, 1, 1, 1, 1, 1, 1}
		maxOctave 		= {7, 7, 7, 7, 7, 7, 7, 7, 7, 7}
		minVolume 		= 0
		maxVolume 		= {15, 15, 15, 15, 127, 127, 127, 127, 127, 127}
		supportsPan 		= 0
		minNote 		= {10, 10, 10, 10, 1, 1, 1, 1, 1, 1}
		maxLoopDepth 		= 2
		supportsPAL 		= 1
		adsrLen			= 5
		adsrMax			= 31
		
	elsif target = TARGET_GBC then
		define("GBC", 1)
		supportedChannels 	= {'A', 'B', 'C', 'D'}
		supportsDutyChange 	= { 3,   3,  -1,   1}
		supportsVolumeChange 	= { 1,   1,   1,   1}
		supportsFM		= { 0,   0,   0,   0}
		supportsWave 		= { 0,   0,   1,   0}
		supportsPCM 		= { 0,   0,   0,   0}
		activeChannels 		= { 0,   0,   0,   0}
		supportsHwToneEnv	= { 7,   0,   0,   0}
		supportsHwVolEnv	= { 7,   7,   0,   7}
		supportsDetune		= { 1,   1,   0,   0}
		maxTempo 		= 250
		minNote 		= repeat(1, length(supportedChannels))
		minOctave 		= {2, 2, 2, 2}
		maxOctave 		= {7, 7, 7, 7}
		minVolume 		= 0
		maxVolume 		= {15, 15, 3, 15}
		supportsPan 		= 0
		maxLoopDepth 		= 2
		maxWavLength		= 32
		minWavSample 		= 0
		maxWavSample		= 15
		
	elsif target = TARGET_NDS then
		define("NDS", 1)
		supportedChannels 	= {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J'}
		supportsDutyChange 	= { 7,   7,   7,   7,  -1,  -1,  -1,  -1,  -1,  -1}
		supportsVolumeChange 	= { 1,   1,   1,   1,   1,   1,   1,   1,   1,   1}
		supportsFM		= { 0,   0,   0,   0,   0,   0,   0,   0,   0,   0}
		supportsWave 		= { 0,   0,   0,   0,   0,   0,   1,   1,   1,   1}
		supportsPCM 		= { 0,   0,   0,   0,   0,   0,   1,   1,   1,   1}
		activeChannels 		= { 0,   0,   0,   0,   0,   0,   0,   0,   0,   0}
		supportsHwToneEnv	= { 0,   0,   0,   0,   0,   0,   0,   0,   0,   0}
		supportsHwVolEnv	= { 0,   0,   0,   0,   0,   0,   0,   0,   0,   0}
		supportsDetune 		= { 1,   1,   1,   1,   1,   1,   1,   1,   1,   1}
		maxTempo 		= 250
		minNote 		= repeat(1, length(supportedChannels))
		minOctave 		= {2, 2, 2, 2, 2, 2, 2, 2, 2, 2}
		maxOctave 		= {7, 7, 7, 7, 7, 7, 7, 7, 7, 7}
		minVolume 		= 0
		maxVolume 		= {127, 127, 127, 127, 127, 127, 127, 127, 127, 127}
		supportsPan 		= 1
		maxLoopDepth 		= 4
		maxWavLength		= 262144
		minWavSample 		= -128
		maxWavSample		= 127
		
	elsif target = TARGET_XGS then
		define("XGS", 1)
		supportedChannels 	= {'A', 'B', 'C'}
		supportsDutyChange 	= { 3,   3,   3}
		supportsVolumeChange 	= { 1,   1,   1}
		supportsFM		= { 0,   0,   0}
		supportsWave 		= { 0,   0,   0}
		supportsPCM 		= { 0,   0,   0}
		activeChannels 		= { 0,   0,   0}
		supportsHwToneEnv	= { 0,   0,   0}
		supportsHwVolEnv	= { 0,   0,   0}
		supportsDetune		= { 0,   0,   0}
		maxTempo 		= 250
		minNote 		= repeat(1, length(supportedChannels))
		minOctave 		= {2, 2, 2}
		maxOctave 		= {7, 7, 7}
		minVolume 		= 0
		maxVolume 		= {15, 15, 15}
		supportsPan 		= 1
		maxLoopDepth 		= 2
		supportsPAL		= 1
		useFractionalDelays	= 0
		
	else
		ERROR("No target specified", -1)
	end if

	machineVolLimit = maxVolume
	rest = (max(maxOctave) + 1 - min(minOctave)) * 12
	
	n = length(fileNames[1])
	while n > 0 and fileNames[1][n] != '.' do
		n -= 1
	end while
	if n = 0 then
		shortFilename = fileNames[1]
		fileNames[1] &= ".mml"
	else
		shortFilename = fileNames[1][1..n - 1]
	end if
	if length(fileNames) > 1 then
		shortFilename = fileNames[2]
		n = length(fileNames[2])
		while n > 0 and fileNames[2][n] != '.' do
			n -= 1
		end while
		if n != 0 then
			shortFilename = fileNames[2][1..n - 1]
			writeVGM = equal(lower(fileNames[2][n..length(fileNames[2])]), ".vgm")
		end if
	end if
	inFile = open(fileNames[1], "rb")
	if inFile = -1 then
		ERROR("Unable to open file: " & fileNames[1], -1)
	end if
	
	n = 0
	fileData = ""
	while n != -1 do
		n = getc(inFile)
		if n != -1 then
			fileData &= n
		end if
	end while
	close(inFile)
	fileDataPos = 1
	fileData &= -1
	
	songs 		= repeat(0, 100)
	songLen 	= repeat(0, 100)
	songs[1] 	= repeat({}, length(supportedChannels))
	songLen[1] 	= repeat(0, length(supportedChannels))

	songLoopLen 	= songLen
	songNum 	= 1
	lineNum 	= 1
	prevLine 	= 1
	updateFreq 	= 60.0
	tie 		= 0
	slur 		= 0
	octaveRev 	= 1
	enRev 		= 0
	lastWasChannelSelect = 0
	
	dontCompile	= {0}
	hasElse		= {0}
	macros 		= assoc_create()
	volumeMacros 	= assoc_create()
	panMacros 	= assoc_create()
	volumeSlides 	= assoc_create()
	dutyMacros 	= assoc_create()
	pitchMacros 	= assoc_create()
	arpeggios 	= assoc_create()
	vibratos 	= assoc_create()
	portamentos 	= assoc_create()
	waveforms	= assoc_create()
	pcms		= assoc_create()
	adsrs		= assoc_create()
	mods		= assoc_create()
	callbacks 	= {}
	loopStack 	= repeat({}, length(supportedChannels))
	currentLength 	= repeat(8, length(supportedChannels))
	currentOctave 	= repeat(4, length(supportedChannels))
	currentTempo 	= repeat(125, length(supportedChannels))
	currentCutoff 	= repeat(8, length(supportedChannels))
	currentVolume 	= floor(maxVolume / 2)
	loopPoint 	= repeat(-1, length(supportedChannels))
	currentNote 	= repeat({}, length(supportedChannels))
	lastSetLength 	= repeat(0, length(supportedChannels))
	pendingOctChange = repeat(0, length(supportedChannels))
	hasAnyNote    	= repeat(0, length(supportedChannels))
	tuple		= repeat(0, length(supportedChannels))
	usesEN		= {0, 0}
	
	while 1 do
		bypass = 0
		
		n = getch()
		if n = -1 then
			exit
		end if
		
		n2 = n
		
		if n = 10 then
			lineNum += 1
		end if
		
		if lineNum > prevLine then
			activeChannels = repeat(0, length(supportedChannels))
			prevLine = lineNum
		end if
		
		if dontCompile[1] then
			if n = '#' then
				s = get_string()
				if equal(s, "IFDEF") then
					s = get_string_until("&| "&{13, 10})
					l = is_defined(s)
					while 1 do
						s = get_string_in_range("|&")
						if equal(s, "&") then
							s = get_string_until("&| "&{13, 10})
							l = and_bits(l, is_defined(s))
						elsif equal(s, "|") then
							s = get_string_until("&| "&{13, 10})
							l = or_bits(l, is_defined(s))
						else
							exit
						end if
					end while
					dontCompile = or_bits((not l), dontCompile[1]) & dontCompile
					hasElse = 0 & hasElse
					
				elsif equal(s, "IFNDEF") then
					s = get_string_until("&| "&{13, 10})
					l = is_defined(s)
					while 1 do
						s = get_string_in_range("|&")
						if equal(s, "&") then
							s = get_string_until("&| "&{13, 10})
							l = or_bits(l, is_defined(s))
						elsif equal(s, "|") then
							s = get_string_until("&| "&{13, 10})
							l = and_bits(l, is_defined(s))
						else
							exit
						end if
					end while
					dontCompile = or_bits(l, dontCompile[1]) & dontCompile
					hasElse = 0 & hasElse
					
				elsif equal(s, "ELSIFDEF") then
					s = get_string_until("&| "&{13, 10})
					l = is_defined(s)
					while 1 do
						s = get_string_in_range("|&")
						if equal(s, "&") then
							s = get_string_until("&| "&{13, 10})
							l = and_bits(l, is_defined(s))
						elsif equal(s, "|") then
							s = get_string_until("&| "&{13, 10})
							l = or_bits(l, is_defined(s))
						else
							exit
						end if
					end while
					if length(dontCompile) > 1 then
						if not hasElse[1] then
							if and_bits(dontCompile[1], 2) != 2 then
								dontCompile[1] = or_bits((not l), dontCompile[2])
							end if
						else
							ERROR("ELSIFDEF found after ELSE", lineNum)
						end if
					else
						ERROR("ELSIFDEF with no matching IFDEF", lineNum)
					end if
					
				elsif equal(s, "ELSE") then
					if length(dontCompile) > 1 then
						if not hasElse[1] then
							if and_bits(dontCompile[1], 2) != 2 then
								dontCompile[1] = or_bits((not dontCompile[1]), dontCompile[2])
							end if
							hasElse[1] = 1
						else
							ERROR("Only one ELSE allowed per IFDEF", lineNum)
						end if
					else
						ERROR("ELSE with no matching IFDEF", lineNum)
					end if

				elsif equal(s, "ENDIF") then
					if length(dontCompile) > 1 then
						dontCompile = dontCompile[2..length(dontCompile)]
						hasElse = hasElse[2..length(hasElse)]
					else
						ERROR("ENDIF with no matching IFDEF", lineNum)
					end if
				end if
			end if
		else
		
			if n = '#' then
				write_note(1)
				s = get_string()
				if equal(s, "TITLE") then
					songTitle = get_string_until({13, 10})

				elsif equal(s, "TUNE") then
					tune = 1
					
				elsif equal(s, "COMPOSER") then
					songComposer = get_string_until({13, 10})

				elsif equal(s, "PROGRAMER") or equal(s, "PROGRAMMER") then
					songProgrammer = get_string_until({13, 10})

				elsif equal(s, "ERROR") then
					skip_whitespace()
					if getch() = '\"' then
						s = get_string_until({'\"'})
						if getch() = '\"' then
							ERROR(s, lineNum)
						else
							ERROR("Malformed #ERROR, missing ending \"", lineNum)
						end if
					else
						ERROR("Malformed #ERROR, missing starting \"", lineNum)
					end if

				elsif equal(s, "WARNING") then
					skip_whitespace()
					if getch() = '\"' then
						s = get_string_until({'\"'})
						if getch() = '\"' then
							WARNING(s, lineNum)
						else
							ERROR("Malformed #WARNING, missing ending \"", lineNum)
						end if
					else
						ERROR("Malformed #WARNING, missing starting \"", lineNum)
					end if
					
				elsif equal(s, "INCLUDE") then
					skip_whitespace()
					if getch() = '\"' then
						s = get_string_until({'\"'})
						if getch() = '\"' then
							inFile = open(s, "rb")
							if inFile = -1 then
								ERROR("Unable to open file: " & s, lineNum)
							end if
							m = 0
							s = ""
							while m != -1 do
								m = getc(inFile)
								if m != -1 then
									s &= m
								end if
							end while
							close(inFile)
							if fileDataPos <= length(fileData) then
								fileData = fileData[1..fileDataPos - 1] & s & fileData[fileDataPos..length(fileData)]
							else
								fileData &= s
							end if
						else
							ERROR("Malformed #INCLUDE, missing ending \"", lineNum)
						end if
					else
						ERROR("Malformed #INCLUDE, missing starting \"", lineNum)
					end if
					
				elsif equal(s, "PAL") then
					if supportsPAL then
						updateFreq = 50.0
					end if

				elsif equal(s, "NTSC") then
					updateFreq = 60.0

				elsif equal(s, "IFDEF") then
					s = get_string_until("&| "&{13, 10})
					l = is_defined(s)
					while 1 do
						s = get_string_in_range("|&")
						if equal(s, "&") then
							s = get_string_until("&| "&{13, 10})
							l = and_bits(l, is_defined(s))
						elsif equal(s, "|") then
							s = get_string_until("&| "&{13, 10})
							l = or_bits(l, is_defined(s))
						else
							exit
						end if
					end while
					dontCompile = or_bits((not l), dontCompile[1]) & dontCompile
					hasElse = 0 & hasElse

				elsif equal(s, "IFNDEF") then
					s = get_string_until("&| "&{13, 10})
					l = is_defined(s)
					while 1 do
						s = get_string_in_range("|&")
						if equal(s, "&") then
							s = get_string_until("&| "&{13, 10})
							l = or_bits(l, is_defined(s))
						elsif equal(s, "|") then
							s = get_string_until("&| "&{13, 10})
							l = and_bits(l, is_defined(s))
						else
							exit
						end if
					end while
					dontCompile = or_bits(l, dontCompile[1]) & dontCompile
					hasElse = 0 & hasElse

				elsif equal(s, "ELSIFDEF") then
					if length(dontCompile) > 1 then
						if not hasElse[1] then
							s = get_string_until({13, 10})
							dontCompile[1] = 2
						else
							ERROR("ELSIFDEF found after ELSE", lineNum)
						end if
					else
						ERROR("ELSIFDEF with no matching IFDEF", lineNum)
					end if
					
				elsif equal(s, "ELSE") then
					if length(dontCompile) > 1 then
						if not hasElse[1] then
							dontCompile[1] = 1 
							hasElse[1] = 1
						else
							ERROR("Only one ELSE allowed per IFDEF", lineNum)
						end if
					else
						ERROR("ELSE with no matching IFDEF", lineNum)
					end if

				elsif equal(s, "ENDIF") then
					if length(dontCompile) > 1 then
						dontCompile = dontCompile[2..length(dontCompile)]
						hasElse = hasElse[2..length(hasElse)]
					else
						ERROR("ENDIF with no matching IFDEF", lineNum)
					end if

				elsif equal(s, "SONG") then
					s = get_string()
					o = value(s)
					if o[1] = GET_SUCCESS then
						if integer(o[2]) and o[2] > 1 and o[2] < 100 then
							if integer(songs[o[2]]) then
								m = -1
								l = 0
								for i = 1 to length(supportedChannels) do
									if songLen[songNum][i] then
										if m = -1 then
											m = songLen[songNum][i]
										elsif songLen[songNum][i] != m then
											WARNING("Mismatch in length between channels", -1)
											exit
										end if
									end if
									songLoopLen[songNum][i] = songLen[songNum][i] - songLoopLen[songNum][i]
									if loopPoint[i] = -1 then
										songs[songNum][i] &= CMD_END
									else
										if not hasAnyNote[i] then
											songs[songNum][i] &= CMD_END
										else
											songs[songNum][i] &= {CMD_JMP, and_bits(loopPoint[i], #FF), floor(loopPoint[i] / #100)}
										end if
									end if
									loopPoint[i] = -1
								end for
								songNum = o[2]
								songs[songNum] = repeat({}, length(supportedChannels))
								songLen[songNum] = repeat(0, length(supportedChannels))
								hasAnyNote = repeat(0, length(supportedChannels))
								songLoopLen[songNum] = songLen[songNum]
								for i = 1 to length(loopStack) do
									if length(loopStack[i]) then
										ERROR("Open [ on channel " & supportedChannels[i], lineNum)
									end if
								end for
							else
								ERROR("Song " & s & " already defined", lineNum)
							end if
						else
							ERROR("Bad song number: " & s, lineNum)
						end if
					else
						ERROR("Bad song number: " & s, lineNum)
					end if

				elsif equal(s, "UNIFORM-VOLUME") then
					s = get_string()
					o = value(s)
					if o[1] = GET_SUCCESS then
						if o[2] > 1 then
							maxVolume = repeat(o[2], length(supportedChannels))
							if o[2] > 255 then
								WARNING("Very large max volume specified: " & s, lineNum)
							end if
						else
							ERROR("Volume must be >= 1: " & s, lineNum)
						end if
					else
						ERROR("Expected a positive number: " & s, lineNum)
					end if

				elsif equal(s, "EN-REV") then
					s = get_string()
					o = value(s)
					if o[1] = GET_SUCCESS then
						if o[2] = 1 then
							enRev = 1
						elsif o[2] = 0 then
						else
							WARNING("Expected 0 or 1, defaulting to 0: " & s, lineNum)
						end if
					else
						ERROR("Expected 0 or 1: " & s, lineNum)
					end if

				elsif equal(s, "OCTAVE-REV") then
					s = get_string()
					o = value(s)
					if o[1] = GET_SUCCESS then
						if o[2] = 1 then
							octaveRev = -1
						elsif o[2] = 0 then
						else
							WARNING("Expected 0 or 1, defaulting to 0: " & s, lineNum)
						end if
					else
						ERROR("Expected 0 or 1: " & s, lineNum)
					end if

				elsif equal(s, "AUTO-BANKSWITCH") then
					WARNING("Unsupported command: AUTO-BANKSWITCH", lineNum)
					s = get_string() 

				else
					ERROR("Unknown command: " & s, lineNum)
				end if


			elsif n = '@' then
				write_note(1)
				m = getch()
				if m = '@' then
					s = "@" & get_numeric_string()
				elsif is_numeric(m) then
					ungetch()
					s = get_numeric_string()
				else
					ungetch()
					s = get_alpha_string()
					s &= get_numeric_string()
					--l = getch()
					--if is_alpha(l) then
					--	s = {m, l} & get_numeric_string()
					--else
					--	ungetch()
					--	s = {m} & get_numeric_string()
					--end if
				end if

				if length(s) then
					if s[1] = '@' then
						o = value(substr(s, 2, length(s)))
						if o[1] = GET_SUCCESS then
							--? o[2]
							idx = assoc_find_key(dutyMacros, o[2]) 
							numChannels = 0
							for i = 1 to length(supportedChannels) do
								if activeChannels[i] then
									numChannels += 1
									if supportsDutyChange[i] then
										songs[songNum][i] &= {CMD_DUTMAC, idx}
										dutyMacros = assoc_reference(dutyMacros, o[2])
									else
										WARNING("Unsupported command for channel " &
											supportedChannels[i] & ": @@", lineNum)
									end if
								end if
							end for
							if numChannels = 0 then
								WARNING("Use of @@ with no channels active", lineNum)
							end if
						else
							ERROR("Expected a number: " & s, lineNum)
						end if
					else
						o = value(s)
						-- Duty cycle macro definition
						if o[1] = GET_SUCCESS then
							idx = assoc_find_key(dutyMacros, o[2]) 
							if sum(activeChannels) = 0 then
								if idx <= 0 then
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											if length(t[2]) or length(t[3]) then
												dutyMacros = assoc_append(dutyMacros, o[2], t)
											else
												ERROR("Empty sequence", lineNum)
											end if
										else
											ERROR("Syntax error", lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								numChannels = 0
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										numChannels += 1
										if o[2] >= 0 and o[2] <= supportsDutyChange[i] then
											songs[songNum][i] &= or_bits(CMD_DUTY, o[2])
										else
											if supportsDutyChange[i] = -1 then
												WARNING("Unsupported command for channel "& 
													supportedChannels[i] & ": @", lineNum)
											else
												ERROR("@ out of range: " & s, lineNum)
											end if
										end if
									end if
								end for
								if numChannels = 0 then
									WARNING("Use of @ with no channels active", lineNum)
								end if
							end if

						-- Pan macro definition
						elsif equal(substr(s, 1, 2), "CS") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								idx = assoc_find_key(panMacros, o[2])
								if idx <= 0 then
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											if length(t[2]) or length(t[3]) then
												if in_range(t[2] & t[3], -63, 63) then
													panMacros = assoc_append(panMacros, o[2], t)
												else
													ERROR("Value of out range (allowed: -63-63): " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Empty sequence", lineNum)
											end if
										else
											ERROR("Bad CS: " & t, lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if			
							
						-- Arpeggio definition
						elsif equal(substr(s, 1, 2), "EN") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								idx = assoc_find_key(arpeggios, o[2])
								if idx <= 0 then
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											if length(t[2]) or length(t[3]) then
												if in_range(t[2] & t[3], -63, 63) then
													--arpeggios[2] = append(arpeggios[2], t)
													arpeggios = assoc_append(arpeggios, o[2], t)
												else
													ERROR("Value of out range (allowed: -63-63): " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Empty sequence", lineNum)
											end if
										else
											ERROR("Bad EN: " & t, lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if						

						-- Vibrato definition
						elsif equal(substr(s, 1, 2), "MP") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								idx = assoc_find_key(vibratos, o[2])
								if idx <= 0 then
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											if length(t[2]) = 3 and length(t[3]) = 0 then
												if in_range(t[2], {0, 1, 0}, {127, 127, 63}) then
													--vibratos[2] = append(vibratos[2], t)
													vibratos = assoc_append(vibratos, o[2], t)
												else
													ERROR("Value of out range: " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Bad MP: " & sprint_list(t), lineNum)
											end if
										else
											ERROR("Bad MP: " & t, lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if						

						-- Amplitude/frequqncy modulator
						elsif equal(substr(s, 1, 3), "MOD") then
							o = value(substr(s, 4, length(s)))
							if o[1] = GET_SUCCESS then
								idx = assoc_find_key(mods, o[2])
								if idx <= 0 then
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											if length(t[2]) = 3 and length(t[3]) = 0 then
												if in_range(t[2], {0, 0, 0}, {7, 7, 7}) then
													mods = assoc_append(mods, o[2], t)
												else
													ERROR("Value of out range: " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Bad MOD: " & sprint_list(t), lineNum)
											end if
										else
											ERROR("Bad MOD: " & t, lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if						
							
						-- Pitch macro definition
						elsif equal(substr(s, 1, 2), "EP") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								idx = assoc_find_key(pitchMacros, o[2]) 
								if idx <= 0 then
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											pitchMacros = assoc_append(pitchMacros, o[2], t)
										else
											ERROR("Bad EP: " & t, lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if		

						-- Portamento definition
						elsif equal(substr(s, 1, 2), "PT") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								idx = assoc_find_key(portamentos, o[2])
								if idx <= 0 then
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											if length(t[2]) = 2 and length(t[3]) = 0 then
												if in_range(t[2], {0, 1}, {127, 127}) then
													--portamentos[2] = append(portamentos[2], t)
													portamentos = assoc_append(portamentos, o[2], t)
												else
													ERROR("Value out of range: " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Bad PT: " & sprint_list(t), lineNum)
											end if
										else
											ERROR("Bad PT: " & t, lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if		

						-- Waveform definition
						elsif equal(substr(s, 1, 2), "WT") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								idx = assoc_find_key(waveforms, o[2])
								if idx <= 0 then
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if sum(supportsWave) != 0 then
											if t[1] = GET_SUCCESS then
												if length(t[3]) = 0 then
													if in_range(t[2], minWavSample, maxWavSample) then
														if length(t[2]) < maxWavLength then
															WARNING("Padding waveform with zeroes", lineNum)
															t[2] &= repeat(0, maxWavLength - length(t[2]))
														elsif length(t[2]) > maxWavLength then
															WARNING("Truncating waveform", lineNum)
															t[2] = t[2][1..maxWavLength]
														end if
														--waveforms[2] = append(waveforms[2], t[2])
														waveforms = assoc_append(waveforms, o[2], t)
													else
														ERROR("Waveform data out of range: " & sprint_list(t), lineNum)
													end if
												else
													ERROR("Loops not supported in waveform: " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Bad waveform: " & t, lineNum)
											end if
										else
											WARNING("Unsupported command for this target: @WT", lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if		
						
						elsif equal(substr(s, 1, 4), "ADSR") then
							o = value(substr(s, 5, length(s)))
							if o[1] = GET_SUCCESS then
								idx = assoc_find_key(adsrs, o[2])
								if idx <= 0 then
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											if length(t[3]) = 0 then
												if length(t[2]) = adsrLen then
													if in_range(t[2], 0, adsrMax) then
														adsrs = assoc_append(adsrs, o[2], t)
													else
														ERROR("ADSR parameters out of range: " & sprint_list(t), lineNum)
													end if
												else
													ERROR("Bad number of ADSR parameters: " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Bad ADSR: " & sprint_list(t), lineNum)
											end if
										else
											ERROR("Bad ADSR: " & t, lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if
							
						-- Volume slide definition
						elsif equal(substr(s, 1, 2), "VS") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								idx = assoc_find_key(volumeSlides, o[2])
								if idx <= 0 then
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											if length(t[2]) = 3 and length(t[3]) = 0 then
												if in_range(t[2], {0, 1, -63}, {127, 127, 63}) then									
													--volumeSlides[2] = append(volumeSlides[2], t)
													volumeSlides = assoc_append(volumeSlides, o[2], t)
												else
													ERROR("Bad VS: " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Bad VS: " & sprint_list(t), lineNum)
											end if
										else
											ERROR("Bad VS: " & t, lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if		

						-- Volume macro definition
						elsif equal(substr(s, 1, 1), "v") then
							o = value(substr(s, 2, length(s)))
							if o[1] = GET_SUCCESS then
								idx = assoc_find_key(volumeMacros, o[2])
								if sum(activeChannels) = 0 then
									if idx <= 0 then
										t = get_string()
										if equal(t, "=") then
											t = get_list()
											if t[1] = GET_SUCCESS then
												if in_range(t[2], minVolume, maxVolume[length(supportedChannels)]) and
												   in_range(t[3], minVolume, maxVolume[length(supportedChannels)]) then
													--volumeMacros[2] = append(volumeMacros[2], t)
													volumeMacros = assoc_append(volumeMacros, o[2], t)
												else
													ERROR("Value out of range: " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Bad volume macro: " & t, lineNum)
											end if
										else
											ERROR("Expected \'=\'", lineNum)
										end if
									else
										ERROR("Redefinition of @" & s, lineNum)
									end if
								else
									if idx <= 0 then
										ERROR("Undefined macro: @" & s, lineNum)
									else
										for i = 1 to length(supportedChannels) do
											if activeChannels[i] then
												if not is_empty_list(volumeMacros[2][idx]) then
													songs[songNum][i] &= {CMD_VOLMAC, idx}
													volumeMacros = assoc_reference(volumeMacros, o[2])
												end if
											end if
										end for
									end if
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if		


						elsif equal(substr(s, 1, 2), "te") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								write_note(1)
								if sum(activeChannels) = 0 then
									WARNING("Trying to set tone envelope with no active channels", lineNum)
								elsif integer(o[2]) then
									for i = 1 to length(supportedChannels) do
										if activeChannels[i] then
											if supportsHwToneEnv[i] then
												if o[2] >= -supportsHwToneEnv[i] and o[2] <= supportsHwToneEnv[i] then
													songs[songNum][i] &= {CMD_HWTE, o[2]}
												else
													ERROR("Tone envelope number out of range: " & s, lineNum)
												end if
											else
												WARNING("Unsupported command for channel " & supportedChannels[i] & ": @" & s, lineNum)
											end if
										end if
									end for
								else
									ERROR("Bad tone envelope: " & s, lineNum)
								end if
							else
								ERROR("Bad tone envelope: " & s, lineNum)
							end if

						elsif equal(substr(s, 1, 2), "ve") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								write_note(1)
								if sum(activeChannels) = 0 then
									WARNING("Trying to set volume envelope with no active channels", lineNum)
								elsif integer(o[2]) then
									for i = 1 to length(supportedChannels) do
										if activeChannels[i] then
											if supportsHwVolEnv[i] then
												if o[2] >= -supportsHwVolEnv[i] and o[2] <= supportsHwVolEnv[i] then
													songs[songNum][i] &= {CMD_HWVE, o[2]}
												else
													ERROR("Volume envelope number out of range: " & s, lineNum)
												end if
											else
												WARNING("Unsupported command for channel " & supportedChannels[i] & ": @" & s, lineNum)
											end if
										end if
									end for
								else
									ERROR("Bad volume envelope: " & s, lineNum)
								end if
							else
								ERROR("Bad volume envelope: " & s, lineNum)
							end if
							
						elsif equal(substr(s, 1, 1), "q") then
							o = value(substr(s, 2, length(s)))
							if o[1] = GET_SUCCESS then
								write_note(1)
								if sum(activeChannels) = 0 then
									WARNING("Trying to set cutoff with no active channels", lineNum)
								elsif integer(o[2]) and o[2] >= 0 and o[2] <= 15 then
									for i = 1 to length(supportedChannels) do
										if activeChannels[i] then
											currentCutoff[i] = -o[2]
										end if
									end for
								else
									ERROR("Bad cutoff: " & s, lineNum)
								end if
							else
								ERROR("Bad cutoff: " & s, lineNum)
							end if
				
						else
							ERROR("Syntax error: @" & s, lineNum)
						end if
					end if
				else
					ERROR("Unexpected end of file", lineNum)
				end if

			-- Macro definition/invokation
			elsif n = '$' then
				write_note(1)
				s = get_string_in_range(ALPHANUM)
				skip_whitespace()
				m = getch()
				if length(s) then
					-- Macro definition
					if m = '{' then
						idx = find(s, macros[1])
						if idx <= 0 then
							if sum(activeChannels) = 0 then
								macro = {}
								while n != ';' do
									n = getch()
									if n = '%' then
										t = get_numeric_string()
										o = value(t)
										if o[1] = GET_SUCCESS then
											macro = append(macro, {1, o[2]})
										else
											ERROR("Syntax error: " & t, lineNum)
										end if
										n = getch()
										if n != '%' then
											ERROR("Missing %%", lineNum)
										end if
									elsif n = '}' or n = -1 then
										exit
									elsif n = 10 then
										lineNum += 1
									elsif n != 13 then
										macro = append(macro, {0, n})
									end if
								end while
							else
								ERROR("Macro definitions are not allowed while channels are active", lineNum)
							end if
							macros[1] = append(macros[1], s)
							macros[2] = append(macros[2], macro)
						else
							ERROR("Macro already defined: " & s, lineNum)
						end if

					-- Macro invokation
					elsif m = '(' then
						macro = {}
						while 1 do
							t = get_string_until(",)\t " & {13,10})
							macro = append(macro, t)
							skip_whitespace()
							n = getch()
							if n = -1 or n = ')' then
								exit
							end if
						end while
						u = ""
						idx = find(s, macros[1])
						if idx > 0 then
							for i = 1 to length(macros[2][idx]) do
								if macros[2][idx][i][1] = 0 then
									u &= macros[2][idx][i][2]
								else
									if macros[2][idx][i][2] > 0 and
									   macros[2][idx][i][2] <= length(macro) then
										u &= macro[macros[2][idx][i][2]]
									else
										ERROR(sprintf("Macro parameter enum out of range: %d", 
										      macros[2][idx][i][2]), lineNum)
									end if
								end if
							end for
							if fileDataPos <= length(fileData) then
								fileData = fileData[1..fileDataPos-1] & u & fileData[fileDataPos..length(fileData)]
							else
								fileData &= u
							end if

							numChannels = 0
							for i = 1 to length(supportedChannels) do
								if activeChannels[i] then
									numChannels += 1
								end if
							end for
							if numChannels = 0 then
								ERROR("Trying to invoke a macro with no channels active", lineNum)
							end if
						else
							ERROR("Undefined macro: " & s, lineNum)
						end if
					else
						ERROR("Syntax error: " & m, lineNum)
					end if
				else
					ERROR("Expected an identifier following $", lineNum)
				end if

			-- Callback
			elsif n = '!' then
				write_note(1)
				s = get_string_until("(\t " & {13,10})
				skip_whitespace()
				n = getch()
				if n = '(' then
					idx = find(s, callbacks)
					if idx <= 0 then
						callbacks = append(callbacks, s)
						idx = length(callbacks)
					end if

					t = get_string_until(")\t " & {13,10})
					o = value(t)
					if o[1] = GET_SUCCESS then
						if o[1] = 0 then
							-- Never (Off)
							numChannels = 0
							for i = 1 to length(supportedChannels) do
								if activeChannels[i] then
									numChannels += 1
									songs[songNum][i] &= CMD_CBOFF
								end if
							end for
							if numChannels = 0 then
								ERROR("Trying to set a callback with no channels active", lineNum)
							end if
						elsif o[1] = 1 then
							-- Once
							numChannels = 0
							for i = 1 to length(supportedChannels) do
								if activeChannels[i] then
									numChannels += 1
									songs[songNum][i] &= {CMD_CBONCE, idx - 1}
								end if
							end for
							if numChannels = 0 then
								ERROR("Trying to set a callback with no channels active", lineNum)
							end if
						else
							ERROR("Bad callback frequency: " & s, lineNum)
						end if
					elsif equal(t, "EVERY-NOTE") then
						numChannels = 0
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								numChannels += 1
								songs[songNum][i] &= {CMD_CBEVNT, idx - 1}
							end if
						end for
						if numChannels = 0 then
							ERROR("Trying to set a callback with no channels active", lineNum)
						end if

					elsif equal(t, "EVERY-VOL-CHANGE") then
						numChannels = 0
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								numChannels += 1
								songs[songNum][i] &= {CMD_CBEVVC, idx - 1}
							end if
						end for
						if numChannels = 0 then
							ERROR("Trying to set a callback with no channels active", lineNum)
						end if

					elsif equal(t, "EVERY-VOL-MIN") then
						numChannels = 0
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								numChannels += 1
								songs[songNum][i] &= {CMD_CBEVVM, idx - 1}
							end if
						end for
						if numChannels = 0 then
							ERROR("Trying to set a callback with no channels active", lineNum)
						end if

					elsif equal(t, "EVERY-OCT-CHANGE") then
						numChannels = 0
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								numChannels += 1
								songs[songNum][i] &= {CMD_CBEVOC, idx - 1}
							end if
						end for
						if numChannels = 0 then
							ERROR("Trying to set a callback with no channels active", lineNum)
						end if

					else
						ERROR("Bad callback frequency: " & s, lineNum)
					end if
					n = getch()
					if n != ')' then
						ERROR("Expected ): " & n, lineNum)
					end if
				else
					ERROR("Expected (: " & n, lineNum)
				end if

			-- Multiline comment
			elsif n = '/' then
				write_note(1)
				m = getch()
				if m = '*' then
					while n != -1 do
						n = getch()
						if n = '*' then
							m = getch()
							if m = '/' then
								exit
							else
								ungetch()
							end if
						elsif n = 10 then
							lineNum += 1
						end if
					end while
				else
					ERROR("Syntax error: " & n, lineNum)
				end if

			elsif n = '[' then
				write_note(1)
				for i = 1 to length(supportedChannels) do
					if activeChannels[i] then
						loopStack[i] = push(loopStack[i], {length(songs[songNum][i]) + 2, songLen[songNum][i]})
						if length(loopStack[i]) > maxLoopDepth then
							ERROR(sprintf("Maximum nesting level for loops is %d", maxLoopDepth), lineNum)
						end if
						songs[songNum][i] &= {CMD_LOPCNT, 0}
					end if
				end for

			elsif n = ']' then
				write_note(1)
				t = get_numeric_string()
				o = value(t)
				for i = 1 to length(supportedChannels) do
					if activeChannels[i] then
						s = pop(loopStack[i])
						loopStack[i] = s[2]
						if s[1][1] != 0 then
							if o[1] = GET_SUCCESS then
								if o[2] > 0 then
									songs[songNum][i][s[1][1]] = o[2]	-- Loop count
									songs[songNum][i] &= {CMD_DJNZ, and_bits(s[1][1], #FF), floor(s[1][1] / #100)}
									songLen[songNum][i] += (songLen[songNum][i] - s[1][2]) * (o[2] - 1)
								else
									ERROR("Bad loop count: " & t, lineNum)
								end if
							else
								? activeChannels
								ERROR("Expected a loop count: " & t, lineNum)
							end if
						else
							ERROR("Use of ] with no matching [ on channel " & supportedChannels[i], lineNum)
						end if
					end if
				end for

			-- Octave up/down		
			elsif n = '>' or n = '<' then
				write_note(1)
				m = 1
				if n = '<' then
					m = -1
				end if
				if sum(activeChannels) = 0 then
					WARNING("Trying to change octave with no active channels: " & n, lineNum)
				else
					for i = 1 to length(supportedChannels) do
						if activeChannels[i] then
							if currentOctave[i] + m >= minOctave[i] and
							   currentOctave[i] + m <= maxOctave[i] then
								currentOctave[i] += m
								pendingOctChange[i] = m
							else
								ERROR(sprintf("Octave out of range: " & n & " (%d)", currentOctave[i] + m), lineNum)
							end if
						end if
					end for
				end if

			elsif n = 'l' then
				write_note(1)
				s = get_numeric_string()
				o = value(s)
				if o[1] = GET_SUCCESS then
					if sum(activeChannels) = 0 then
						WARNING("Trying to set length with no active channels", lineNum)
					elsif integer(o[2]) and find(o[2], SUPPORTED_LENGTHS) > 0 then
						numChannels = 0
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								numChannels += 1
								currentLength[i] = floor(power(2, find(o[2], SUPPORTED_LENGTHS) - 1))
							end if
						end for
						if numChannels = 0 then
							WARNING("Trying to set length with no active channels", lineNum)
						end if
					else
						ERROR("Bad length: " & s, lineNum)
					end if
				else
					ERROR("Bad length: " & s, lineNum)
				end if

			-- Set octave
			elsif n = 'o' then
				write_note(0)
				s = get_numeric_string()
				o = value(s)
				if o[1] = GET_SUCCESS then
					if sum(activeChannels) = 0 then
						WARNING("Trying to set octave with no active channels", lineNum)
					elsif integer(o[2]) then 
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								if o[2] >= minOctave[i] and o[2] <= maxOctave[i] then
									currentOctave[i] = o[2]
									songs[songNum][i] &= or_bits(CMD_OCTAVE, currentOctave[i] - minOctave[i])
								else
									ERROR("Octave out of range", lineNum)
								end if
							end if
						end for
					else
						ERROR("Bad octave: " & s, lineNum)
					end if
				else
					ERROR("Bad octave: " & s, lineNum)
				end if

			-- Set cutoff
			elsif n = 'q' then
				write_note(1)
				s = get_numeric_string()
				o = value(s)
				if o[1] = GET_SUCCESS then
					if sum(activeChannels) = 0 then
						WARNING("Trying to set cutoff with no active channels", lineNum)
					elsif integer(o[2]) and o[2] >= 0 and o[2] <= 8 then
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								currentCutoff[i] = o[2]
							end if
						end for
					else
						ERROR("Bad cutoff: " & s, lineNum)
					end if
				else
					ERROR("Bad cutoff: " & s, lineNum)
				end if

			-- Set tempo
			elsif n = 't' then
				write_note(1)
				s = get_numeric_string()
				o = value(s)
				if o[1] = GET_SUCCESS then
					if sum(activeChannels) = 0 then
						WARNING("Trying to set tempo with no active channels", lineNum)
					elsif integer(o[2]) and o[2] >= 0 and o[2] < maxTempo then
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								currentTempo[i] = o[2]
							end if
						end for
					else
						ERROR("Bad tempo: " & s, lineNum)
					end if
				else
					ERROR("Bad tempo: " & s, lineNum)
				end if

			-- Set volume
			elsif n = 'v' then
				write_note(1)
				s = get_numeric_string()
				if length(s) then
					o = value(s)
				else
					m = getch()
					if m = '+' then
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								if currentVolume[i] < maxVolume[i] then
									o = {GET_SUCCESS, currentVolume[i] + 1}
								else
									WARNING("Volume too high on channel " & supportedChannels[i], lineNum)
								end if
							end if
						end for
					elsif m = '-' then
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								if currentVolume[i] > minVolume then
									o = {GET_SUCCESS, currentVolume[i] -  1}
								else
									WARNING("Volume too low on channel " & supportedChannels[i], lineNum)
								end if
							end if
						end for
					else
						ERROR("Expected +, - or a number: " & m, lineNum)
					end if
				end if

				if o[1] = GET_SUCCESS then
					if sum(activeChannels) = 0 then
						WARNING("Trying to set volume with no active channels", lineNum)
					elsif integer(o[2]) and o[2] >= minVolume then
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								if o[2] <= maxVolume[i] then
									currentVolume[i] = o[2]
									if supportsVolumeChange[i] then
										l = floor((currentVolume[i] / maxVolume[i]) * machineVolLimit[i])
										if l <= #0F then
											songs[songNum][i] &= or_bits(CMD_VOL2, l)
										else
											songs[songNum][i] &= {CMD_VOLSET, l}
										end if
									else
										-- TODO: Handle this case (e.g. NES triangle channel)
									end if
								else
									ERROR("Bad volume: " & s, lineNum)
								end if
							end if
						end for
					else
						ERROR("Bad volume: " & s, lineNum)
					end if
				else
					ERROR("Bad volume: " & s, lineNum)
				end if

			-- Single line comment
			elsif n = ';' then
				write_note(1)
				while n != -1 do
					n = getch()
					if n = 10 then
						lineNum += 1
						exit
					end if
				end while

			elsif n = 'L' then
				write_note(1)
				if sum(activeChannels) = 0 then
					WARNING("Trying to set a loop point with no active channels", lineNum)
				else
					for i = 1 to length(supportedChannels) do
						if activeChannels[i] then
							if loopPoint[i] = -1 then
								loopPoint[i] = length(songs[songNum][i])
								songLoopLen[songNum][i] = songLen[songNum][i]
							else
								ERROR("Loop point already defined for channel " &
								      supportedChannels[i], lineNum)
							end if
						end if
					end for
				end if

			-- Reads notes (e.g. c d+ a+4 g-1.. e1^8^32 f&f16&f32)
			elsif find(n, NOTES) > 0 then
				if not slur then
					write_note(0)
				end if

				dot 	= 0
				tie 	= 0
				slur 	= 0
				hasTie 	= 0
				hasSlur = 0
				firstNote = -1
				len = -1
				
				l = 0
				while n != -1 do
					if find(n, NOTES) > 0 then
						if l > 0 and not slur then
							ungetch()
							exit
						end if
						note = find(n, NOTES)
					elsif n = '&' then
						if hasTie then
							ERROR("Trying to use & and ^ in same expression", lineNum)
						end if
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] and sequence(tuple[i]) then
								WARNING("Command ignored inside tuple: &", lineNum)
							end if
						end for
						hasSlur = 1
						slur = 1
						note = -1
					elsif n = '^' then
						if hasSlur then
							ERROR("Trying to use & and ^ in same expression", lineNum)
						end if
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] and sequence(tuple[i]) then
								WARNING("Command ignored inside tuple: ^", lineNum)
							end if
						end for
						hasTie = 1
						tie = 1
					elsif n = '.' then
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] and sequence(tuple[i]) then
								WARNING("Command ignored inside tuple: .", lineNum)
							end if
						end for
						dot = 1
					else
						ungetch()
						exit
					end if

					if (find(n, NOTES) > 0 and n != 'r') or (l > 0 and note != -1 and slur) then
						flatSharp = 0
						len = -1
						m = getch()
						--t = ""
						--while m = '+' or m = '-' do
						--	if m = '+' then
						--		flatSharp += 1
						--		t &= m
						--	elsif m = '-' then
						--		flatSharp -= 1
						--		t &= m
						--	else
						--		exit
						--	end if
						--	m = getch()
						--end while
						--ungetch()
						--if NOTES[remainder(note + flatSharp, 12) + 1] != -2
						--	ERROR("Bad note: " & n & t, lineNum)
						--end if
						if m = '+' then
							flatSharp = 1
							if NOTES[note + flatSharp] != -2 then
								ERROR("Bad note: " & n & "+", lineNum)
							end if
						elsif m = '-' then
							flatSharp = -1
							if note + flatSharp < 1 or NOTES[note + flatSharp] != -2 then
								ERROR("Bad note: " & n & "-", lineNum)
							end if
						else
							ungetch()
						end if
					elsif n = 'r' then
						flatSharp = 0
					end if

					if firstNote = -1 then
						firstNote = note + flatSharp
					else
						if firstNote != note + flatSharp and note != -1 then
							ERROR("Trying to concatenate different notes", lineNum)
						end if
					end if

					if find(n, NOTES) > 0 or tie then
						s = get_numeric_string()
						if length(s) then
							o = value(s)
							if o[1] = GET_SUCCESS then
								if find(o[2], SUPPORTED_LENGTHS) > 0 then
									len = floor(power(2, find(o[2], SUPPORTED_LENGTHS) - 1))
									for i = 1 to length(supportedChannels) do
										if activeChannels[i] and sequence(tuple[i]) then
											WARNING("Note length ignored inside tuple", lineNum)
										end if
									end for
								else
									ERROR("Unsupported length: " & s, lineNum)
								end if
							else
								ERROR("Bad length: " & s, lineNum)
							end if
						end if
					end if
					if len = -1 and tie then
						ERROR("Expected a length", lineNum)
					end if

					numChannels = 0
					for i = 1 to length(supportedChannels) do
						if activeChannels[i] then
							if currentOctave[i] <= minOctave[i] and
							   note < minNote[i] then
								WARNING("The frequency of the note can not be reproduced for this octave: " & NOTES[note], lineNum)
							end if
							numChannels += 1
							if len = -1 then
								len = currentLength[i]
							end if
							if l = 0 then
								if n != 'r' then
									currentNote[i] = {(currentOctave[i] - minOctave[i]) * 12 + note + flatSharp - 1, len}
								else
									currentNote[i] = {rest, len}
								end if
								lastSetLength[i] = len
							else
								if dot then
									len = floor(lastSetLength[i] / 2)
									lastSetLength[i] = len
									if len >= 1 then
										currentNote[i][2] += len
									else
										ERROR("Note length out of range due to dot command", lineNum)
									end if
									dot = 0
								elsif tie then
									currentNote[i][2] += len
									tie = 0
								elsif slur then
									if note != -1 then
										if (currentOctave[i] - minOctave[i]) * 12 + note + flatSharp - 1 = currentNote[i][1] then
											currentNote[i][2] += len
											lastSetLength[i] = len
											slur = 0
										else
											ERROR("Bad note: " & n, lineNum)
										end if
									end if
								else
									ERROR("Bad note: " & n, lineNum)
								end if
							end if
						end if
					end for
					if numChannels = 0 then
						ERROR("Trying to play a note with no active channels", lineNum)
					end if

					n = getch()
					l += 1
				end while
				tie = 0
				slur = 0
				dot = 0
				lastSetLength = repeat(0, length(supportedChannels))

			elsif n = '{' then
				write_note(1)
				if sum(activeChannels) = 0 then
					ERROR("Trying to play a tuple with no active channels", lineNum)
				else
					for i = 1 to length(supportedChannels) do
						if activeChannels[i] then
							if integer(tuple[i]) then
								tuple[i] = {}
							else
								ERROR("Nested tuples are not allowed", lineNum)
							end if
						end if
					end for
				end if
				
			elsif n = '}' then
				write_note(1)
				s = get_numeric_string()
				len = -1
				if length(s) then
					o = value(s)
					if o[1] = GET_SUCCESS then
						if find(o[2], SUPPORTED_LENGTHS) > 0 then
							len = floor(power(2, find(o[2], SUPPORTED_LENGTHS) - 1))
						else
							ERROR("Unsupported length: " & s, lineNum)
						end if
					else
						ERROR("Bad length: " & s, lineNum)
					end if
				end if
				if len = -1 then
					ERROR("No length specified for tuple", lineNum)
				end if
				if sum(activeChannels) = 0 then
					ERROR("Trying to close a tuple with no active channels", lineNum)
				else
					numChannels = 0
					for i = 1 to length(supportedChannels) do
						if activeChannels[i] then
							if sequence(tuple[i]) then
								numChannels += 1
								write_tuple(i, tuple[i], len)
								tuple[i] = 0
							end if
						end if
					end for
					if numChannels = 0 then
						ERROR("Found } with no matching {", lineNum)
					end if
				end if
							
				
			elsif find(n, supportedChannels) > 0 or find(n, "ACDEFMOPRW") > 0 then
				write_note(1)
				
				if n = 'A' then
					m = getch()
					s = {m}
					s &= getch()
					s &= getch()
					if equal(s, "DSR") then
						bypass = 1
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							idx = assoc_find_key(adsrs, o[2])
							if sum(activeChannels) = 0 then
								ERROR("ADSR requires at least one active channel", lineNum)
							elsif idx > 0 then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										if supportsFM[i] then
											songs[songNum][i] &= {CMD_ADSR, idx}
											adsrs = assoc_reference(adsrs, o[2])
										else
											WARNING("Unsupported command for this channel: ADSR", lineNum)
										end if
									end if
								end for
							else
								ERROR("Undefined macro: ADSR" & s, lineNum)
							end if
						else
							ERROR("Expected a number: " & s, lineNum)
						end if
					elsif m = 'M' then
						ungetch()
						ungetch()
						s = get_numeric_string()
						if length(s) then
							bypass = 1
							o = value(s)
							if o[1] = GET_SUCCESS then
								if in_range(o[2], 0, 1) then
									if sum(activeChannels) = 0 then
										ERROR("AM requires at least one active channel", lineNum)
									else
										for i = 1 to length(supportedChannels) do
											if activeChannels[i] then
												if supportsFM[i] then
													songs[songNum][i] &= {CMD_HWAM, o[2]}
												else
													WARNING("Unsupported command for this channel: AM", lineNum)
												end if
											end if
										end for
									end if
								else
									ERROR("AM out of range", lineNum)
								end if
							else
								ERROR("Bad AM: " & s, lineNum)
							end if
						else
							ungetch()
							if find(n, supportedChannels) <= 0 then
								ERROR("Unexpected character: " & n, lineNum)
							end if
						end if						
					else
						ungetch()
						ungetch()
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if						

				elsif n = 'C' then
					m = getch()
					if m = 'S' then
						bypass = 1
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							idx = find(o[2], panMacros[1])
							if sum(activeChannels) = 0 then
								ERROR("CS requires at least one active channel", lineNum)
							elsif idx > 0 then
								if supportsPan then
							--		if sum(activeChannels) > 0 then
									for i = 1 to length(supportedChannels) do
										if activeChannels[i] then
											songs[songNum][i] &= {CMD_PANMAC, idx}
											panMacros = assoc_reference(panMacros, o[2])
										end if
									end for
							--		else
							--			WARNING("Trying to use CS with no channels active", lineNum)
							--		end if
								else
									WARNING("Unsupported command for this target: CS", lineNum)
								end if
							elsif supportsPan then
								ERROR("Undefined macro: CS" & s, lineNum)
							end if
						else
							m = getch()
							t = m & getch()
							if equal(t, "OF") then
								if supportsPan then
									for i = 1 to length(supportedChannels) do
										if activeChannels[i] then
											songs[songNum][i] &= {CMD_PANMAC, 0}
										end if
									end for
								end if
							else
								ERROR("Syntax error: CS" & s, lineNum)
							end if
						end if
					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if						
				
				elsif n = 'D' then
					m = getch()
					if m = '-' or is_numeric(m) then
						bypass = 1
						ungetch()
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							if sum(activeChannels) = 0 then
								ERROR("Detune requires at least one active channel", lineNum)
							else
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										if supportsDetune[i] then
											if supportsFM[i] and target = TARGET_SMD then
												if in_range(o[2], -3, 3) then
													l = o[2]
													if l < 0 then
														l = (-l) + 4
													end if
													songs[songNum][i] &= {CMD_DETUNE, l}
												else
													ERROR("Detune value out of range: " & s, lineNum)
												end if
											else
												if in_range(o[2], -127, 127) then
													songs[songNum][i] &= {CMD_DETUNE, o[2]}
												else
													ERROR("Detune value out of range: " & s, lineNum)
												end if
											end if
										else
											WARNING("Unsupported command for this channel: D", lineNum)
										end if
									end if
								end for
							end if
						else
							ERROR("Expected a number: " & s, lineNum)
						end if
					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if						
						
				elsif n = 'E' then
					m = getch()
					if m = 'N' then
						bypass = 1
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							idx = find(o[2], arpeggios[1])
							if sum(activeChannels) = 0 then
								ERROR("EN requires at least one active channel", lineNum)
							elsif idx > 0 then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										if not is_empty_list(arpeggios[2][idx]) then
											if enRev = 0 then
												songs[songNum][i] &= {CMD_ARPMAC, idx}
												usesEN[1] += 1
											else
												songs[songNum][i] &= {CMD_APMAC2, idx}
												usesEN[2] += 1
											end if
										end if
									end if
								end for
							else
								ERROR("Undefined macro: EN" & s, lineNum)
							end if
						else
							m = getch()
							t = m & getch()
							if equal(t, "OF") then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										if enRev = 0 then
											songs[songNum][i] &= CMD_ARPOFF 
										else
											songs[songNum][i] &= CMD_ARPOFF 
										end if
									end if
								end for
							else
								ERROR("Syntax error: EN" & s, lineNum)
							end if
						end if

					elsif m = 'P' then
						bypass = 1
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							idx = find(o[2], pitchMacros[1])
							if sum(activeChannels) = 0 then
								ERROR("EP requires at least one active channel", lineNum)
							elsif idx > 0 then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										if not is_empty_list(pitchMacros[2][idx]) then
											songs[songNum][i] &= {CMD_SWPMAC, idx}
											pitchMacros = assoc_reference(pitchMacros, o[2])
										end if
									end if
								end for
							else
								ERROR("Undefined macro: EP" & s, lineNum)
							end if
						else
							m = getch()
							t = m & getch()
							if equal(t, "OF") then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then	
										songs[songNum][i] &= {CMD_SWPMAC, 0}
									end if
								end for
							else
								ERROR("Syntax error: EP" & s, lineNum)
							end if
						end if

					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if

				-- FM feedback select
				elsif n = 'F' then
					m = getch()
					if m = 'B' then
						s = get_numeric_string()
						if length(s) then
							bypass = 1
							o = value(s)
							if o[1] = GET_SUCCESS then
								if sum(activeChannels) = 0 then
									ERROR("FB requires at least on active channel", lineNum)
								else
									for i = 1 to length(supportedChannels) do
										if activeChannels[i] then
											if supportsFM[i] then
												if o[2] >=0 and o[2] <= 7 then
													songs[songNum][i] &= or_bits(CMD_FEEDBK, o[2])
												else
													ERROR(sprintf("Feedback value out of range: %d", o[2]), lineNum)
												end if
											else
												WARNING("FB commands on non-FM channels are ignored", lineNum)
											end if
										end if
									end for
								end if
							else
								ERROR("Bad feedback value: " & s, lineNum)
							end if
						else
							ungetch()
							if find(n, supportedChannels) <= 0 then
								ERROR("Unexpected character: " & n, lineNum)
							end if
						end if
					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if					

				-- Vibrato select						
				elsif n = 'M' then
					m = getch()
					if m = 'P' then
						bypass = 1
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							idx = find(o[2], vibratos[1])
							if sum(activeChannels) = 0 then
								ERROR("MP requires at least one active channel", lineNum)
							elsif idx > 0 then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										songs[songNum][i] &= {CMD_VIBMAC, idx}
										vibratos = assoc_reference(vibratos, o[2])
									end if
								end for
							else
								ERROR("Undefined macro: MP" & s, lineNum)
							end if
						else
							m = getch()
							t = m & getch()
							if equal(t, "OF") then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then	
										songs[songNum][i] &= {CMD_VIBMAC, 0}
									end if
								end for
							else
								ERROR("Syntax error: MP" & s, lineNum)
							end if
						end if
					elsif m = 'F' then
						s = get_numeric_string()
						if length(s) then
							bypass = 1
							o = value(s)
							if o[1] = GET_SUCCESS then
								if in_range(o[2], 0, 15) then
									if sum(activeChannels) = 0 then
										ERROR("MF requires at least one active channel", lineNum)
									else
										for i = 1 to length(supportedChannels) do
											if activeChannels[i] then
												if supportsFM[i] then
													songs[songNum][i] &= {CMD_MULT, o[2]}
												else
													WARNING("MF ignored for non-FM channel", lineNum)
												end if
											end if
										end for
									end if
								else
									ERROR("MF out of range: " & s, lineNum)
								end if
							else
								ERROR("Bad MF: " & s , lineNum)
							end if
						else
							ungetch()
							if find(n, supportedChannels) <= 0 then
								ERROR("Unexpected character: " & n, lineNum)
							end if
						end if
					elsif m = 'O' then
						m = getch()
						if m = 'D' then
							s = get_numeric_string()
							if length(s) then
								bypass = 1
								o = value(s)
								if o[1] = GET_SUCCESS then
									idx = find(o[2], mods[1])
									if sum(activeChannels) = 0 then
										ERROR("MOD requires at least one active channel", lineNum)
									elsif idx > 0 then
										for i = 1 to length(supportedChannels) do
											if activeChannels[i] then
												if supportsFM[i] then
													songs[songNum][i] &= {CMD_MODMAC, idx}
													mods = assoc_reference(mods, o[2])
												end if
											end if
										end for
									else
										ERROR("Undefined macro: MOD" & s, lineNum)
									end if
								else
									ERROR("Bad MOD: " & s, lineNum)
								end if
							else
								ungetch()
								ungetch()
								if find(n, supportedChannels) <= 0 then
									ERROR("Unexpected character: " & n, lineNum)
								end if
							end if
						else
							ungetch()
							ungetch()
							if find(n, supportedChannels) <= 0 then
								ERROR("Unexpected character: " & n, lineNum)
							end if
						end if
					elsif is_numeric(m) then
						bypass = 1
						ungetch()
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							if sum(activeChannels) = 0 then
								ERROR("M requires at least one active channel", lineNum)
							elsif in_range(o[2], 0, 15) then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										songs[songNum][i] &= or_bits(CMD_MODE, o[2])
									end if
								end for
							else
								ERROR("Bad mode: " & s, lineNum)
							end if
						else
							ERROR("Bad mode: " & s, lineNum)
						end if
					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if

				-- FM operator select
				elsif n = 'O' then
					m = getch()
					if m = 'P' then
						s = get_numeric_string()
						if length(s) then
							bypass = 1
							o = value(s)
							if o[1] = GET_SUCCESS then
								if in_range(o[2], 0, 4) then
									if sum(activeChannels) = 0 then
										ERROR("OP requires at least one active channel", lineNum)
									else
										for i = 1 to length(supportedChannels) do
											if activeChannels[i] then
												if supportsFM[i] then
													songs[songNum][i] &= or_bits(CMD_OPER, o[2])
												else
													WARNING("OP ignored for non-FM channel", lineNum)
												end if
											end if
										end for
									end if
								else
									ERROR("OP out of range: " & s, lineNum)
								end if
							else
								ERROR("Bad OP: " & s, lineNum)
							end if
						else
							ungetch()
							if find(n, supportedChannels) <= 0 then
								ERROR("Unexpected character: " & n, lineNum)
							end if
						end if
					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if

				-- Portamento select
				elsif n = 'P' then
					m = getch()
					if m = 'T' then
						bypass = 1
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							idx = find(o[2], portamentos[1])
							if sum(activeChannels) = 0 then
								ERROR("PT requires at least one active channel", lineNum)
							elsif idx > 0 then
								-- TODO: set portamento
								WARNING("PT command not yet implemented", lineNum)
							else
								ERROR("Undefined macro: PT" & s, lineNum)
							end if
						else
							m = getch()
							t = m & getch()
							if equal(t, "OF") then
								-- TODO: deactivate portamento
							else
								ERROR("Syntax error: PT" & s, lineNum)
							end if
						end if
					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if

				-- FM rate scale
				elsif n = 'R' then
					m = getch()
					if m = 'S' then
						bypass = 1
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							if sum(activeChannels) = 0 then
								ERROR("RS requires at least one active channel", lineNum)
							elsif in_range(o[2], 0, 3) then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										if supportsFM[i] then
											songs[songNum][i] &= {CMD_RSCALE, o[2]}
										else
											WARNING("Unsupported command for this channel: RS", lineNum)
										end if
									end if
								end for
							else
								ERROR("RS out of range: " & s, lineNum)
							end if
						else
							ERROR("Syntax error: RS" & s, lineNum)
						end if
					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if

				elsif n = 'W' then
					m = getch()
					if m = 'T' then
						bypass = 1
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							idx = find(o[2], waveforms[1])
							if sum(activeChannels) = 0 then
								ERROR("WT requires at least one active channel", lineNum)
							elsif idx > 0 then
								if sum(activeChannels) > 0 then
									for i = 1 to length(supportedChannels) do
										if activeChannels[i] then
											if supportsWave[i] then
												songs[songNum][i] &= {CMD_LDWAVE, idx}
												waveforms = assoc_reference(waveforms, o[2])
											else
												WARNING("Unsupported command for channel " & supportedChannels[i] & ": WT", lineNum)
											end if
										end if
									end for
								else
									WARNING("Trying to use WT with no channels active", lineNum)
								end if
							elsif sum(supportsWave) != 0 then
								ERROR("Undefined macro: WT" & s, lineNum)
							end if
						else
							m = getch()
							t = m & getch()
							if equal(t, "OF") then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										if supportsWave[i] then
											songs[songNum][i] &= {CMD_LDWAVE, 0}
										else
											WARNING("WT commands not supported for channel " & supportedChannels[i], lineNum)
										end if
									end if
								end for
							else
								ERROR("Syntax error: WT" & s, lineNum)
							end if
						end if
					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if
					
				end if
				if not bypass then
					if not lastWasChannelSelect then
						activeChannels = repeat(0, length(supportedChannels))
					end if
					activeChannels[find(n, supportedChannels)] = 1
				end if		
			elsif find(n, "\t " & {13, 10}) > 0 then
				write_note(n = 10)	
			else
				if n = '%' then
					ERROR("Unexpectd character: %%", lineNum)
				else
					ERROR("Unexpected character: " & n, lineNum)
				end if
			end if

			if find(n2, supportedChannels) > 0 then
				lastWasChannelSelect = 1
			else
				lastWasChannelSelect = 0
			end if
		end if
	end while	
	
	write_note(1)
	
	for i = 1 to length(loopStack) do
		if length(loopStack[i]) then
			ERROR("Open [ on channel " & supportedChannels[i], lineNum)
		end if
	end for

	for i = 1 to length(supportedChannels) do
		songLoopLen[songNum][i] = songLen[songNum][i] - songLoopLen[songNum][i]
		if loopPoint[i] = -1 then
			songs[songNum][i] &= CMD_END
		else
			if not hasAnyNote[i] then
				songs[songNum][i] &= CMD_END
			else
				songs[songNum][i] &= {CMD_JMP, and_bits(loopPoint[i], #FF), floor(loopPoint[i] / #100)}
			end if
		end if
	end for
	
	m = -1
	for j = 1 to length(songs) do
		if sequence(songs[j]) then
			for i = 1 to length(supportedChannels) do
				if songLen[j][i] then
					if m = -1 then
						m = songLen[j][i]
					elsif songLen[j][i] != m then
						WARNING(sprintf("Mismatch in length between channels in song %d", j), -1)
						exit
					end if
				end if
			end for
		end if
	end for
	
	if target = TARGET_SMS or target = TARGET_SGG then
		output_sms(writeVGM)
	elsif target = TARGET_SMD then
		output_gen(writeVGM)
	elsif target = TARGET_GBC then
		output_gbc()
	elsif target = TARGET_NDS then
		output_nds()
	end if
	
	puts(1, "\nDone.\n")
end procedure



verbose = 0
warningsAreErrors = 0
target = TARGET_UNKNOWN

procedure show_help(sequence what)
	if length(what) = 0 then
		puts(1, "Usage: xpmc.exe [options] target input [output]\n\n")
		puts(1, "Options:\n")
		puts(1, "\t-h\tShow this information\n") -- / Information about a\n\t                specific command (e.g. -h EP)\n")
		puts(1, "\t-v\tVerbose mode\n")
		puts(1, "\t-w\tTreat warnings as errors\n")
		puts(1,"\nTarget:\n")
		--puts(1,"\t-gba\tGameboy Advance\n")
		--puts(1,"\t-gbc\tGameboy / Gameboy Color\n")
		puts(1,"\t-gen\tSEGA Genesis\n")
		--puts(1,"\t-nds\tNintendo DS\n")
		--puts(1,"\t-sat\tSEGA Saturn\n")
		puts(1,"\t-sgg\tSEGA Game Gear\n")
		puts(1,"\t-sms\tSEGA Master System\n")
	elsif equal(what, "EN") then
		puts(1,"EN:\tArpeggio macro: { num | num}. num is zero or more numbers in\n\tthe range -63-63.\n")
	end if
end procedure


cmdLine = command_line()
if length(cmdLine) <= 2 then
	show_help("")
	abort(0)
end if
for i=3 to length(cmdLine) do
	if cmdLine[i][1] = '-' then
		restOfLine = cmdLine[i][2..length(cmdLine[i])]
		if equal(restOfLine, "v") then
			verbose = 1
		elsif equal(restOfLine, "w") then
			warningsAreErrors = 1
		elsif equal(restOfLine, "h") or equal(restOfLine, "help") then
			if i < length(cmdLine) then
				if equal(cmdLine[i+1], "EN") then
					show_help("EN")
				elsif equal(cmdLine[i+1], "EP") then
					show_help("EP")
				elsif equal(cmdLine[i+1], "MP") then
					show_help("MP")
				elsif equal(cmdLine[i+1], "VS") then
					show_help("VS")
				else
					show_help("")
				end if
			else
				show_help("")
			end if
			exit
		elsif equal(restOfLine, "sms") then
			target = TARGET_SMS
		elsif equal(restOfLine, "sgg") then
			target = TARGET_SGG
		elsif equal(restOfLine, "gen") or equal(restOfLine, "smd") then
			target = TARGET_SMD
		--elsif equal(restOfLine, "gbc") then
		--	target = TARGET_GBC
		--elsif equal(restOfLine, "gba") then
		--	target = TARGET_GBA
		--elsif equal(restOfLine, "nds") then
		--	target = TARGET_NDS
		--elsif equal(restOfLine, "sat") then
		--	target = TARGET_SAT
		--elsif equal(restOfLine, "xgs") then
		--	target = TARGET_XGS
		end if
	else
		compile(cmdLine[i..length(cmdLine)])
		exit
	end if
end for
